import os.path
from pathlib import Path
from typing import cast

import pystac
import pytest
import stac_asset
from pystac import Item
from stac_asset import Config, S3Client

pytestmark = [
    pytest.mark.asyncio,
    pytest.mark.network_access,
]


@pytest.fixture
def asset_href() -> str:
    return "s3://sentinel-cogs/sentinel-s2-l2a-cogs/42/L/TQ/2023/5/S2B_42LTQ_20230524_0_L2A/thumbnail.jpg"


@pytest.fixture
def requester_pays_asset_href() -> str:
    return "s3://usgs-landsat/collection02/level-2/standard/oli-tirs/2023/092/068/LC09_L2SP_092068_20230522_20230524_02_T2/LC09_L2SP_092068_20230522_20230524_02_T2_thumb_small.jpeg"


@pytest.fixture
def requester_pays_item(data_path: Path) -> Item:
    return cast(
        Item,
        pystac.read_file(
            str(data_path / "LC09_L2SP_092068_20230607_20230609_02_T1_SR.json")
        ),
    )


async def test_download(tmp_path: Path, asset_href: str) -> None:
    async with S3Client() as client:
        await client.download_href(asset_href, tmp_path / "out.jpg")

    assert os.path.getsize(tmp_path / "out.jpg") == 6060


async def test_href_exists(asset_href: str) -> None:
    async with S3Client() as client:
        assert await client.href_exists(asset_href)
        assert not await client.href_exists("s3://does-not-exist/not-a-file")


async def test_download_requester_pays_asset(
    tmp_path: Path, requester_pays_asset_href: str
) -> None:
    async with S3Client(requester_pays=True) as client:
        if not await client.has_credentials():
            pytest.skip("aws credentials are invalid or not present")
        await client.download_href(requester_pays_asset_href, tmp_path / "out.jpg")
        assert os.path.getsize(tmp_path / "out.jpg") == 6114


async def test_download_requester_pays_item(
    tmp_path: Path, requester_pays_item: Item
) -> None:
    await stac_asset.download_item(
        requester_pays_item,
        tmp_path,
        config=Config(
            include=["thumbnail"], s3_requester_pays=True, alternate_assets=["s3"]
        ),
    )
    assert (
        os.path.getsize(
            tmp_path / "LC09_L2SP_092068_20230607_20230609_02_T1_thumb_small.jpeg"
        )
        == 19554
    )



"""Blocking interfaces for functions.

These should only be used from fully synchronous code. If you have _any_ async
code in your application, prefer the top-level functions.
"""

import asyncio
from pathlib import Path
from typing import List, Optional

from pystac import Asset, Collection, Item, ItemCollection

from . import _functions
from .client import Client, Clients
from .config import Config
from .types import MessageQueue, PathLikeObject


def download_item(
    item: Item,
    directory: PathLikeObject,
    file_name: Optional[str] = None,
    infer_file_name: bool = True,
    config: Optional[Config] = None,
    messages: Optional[MessageQueue] = None,
    clients: Optional[List[Client]] = None,
    keep_non_downloaded: bool = False,
) -> Item:
    """Downloads an item to the local filesystem, synchronously.

    Args:
        item: The :py:class:`pystac.Item`.
        directory: The output directory that will hold the items and assets.
        file_name: The name of the item file to save. If not provided, will not
            be saved.
        infer_file_name: If ``file_name`` is None, infer the file name from the
            item's id. This argument is unused if ``file_name`` is not None.
        config: The download configuration
        messages: An optional queue to use for progress reporting
        clients: Pre-configured clients to use for access
        keep_non_downloaded: Keep all assets on the item, even if they're not
            downloaded.

    Returns:
        Item: The `~pystac.Item`, with the updated asset hrefs and self href.

    Raises:
        ValueError: Raised if the item doesn't have any assets.
    """
    return asyncio.run(
        _functions.download_item(
            item=item,
            directory=directory,
            file_name=file_name,
            infer_file_name=infer_file_name,
            config=config,
            messages=messages,
            clients=clients,
            keep_non_downloaded=keep_non_downloaded,
        )
    )


def download_collection(
    collection: Collection,
    directory: PathLikeObject,
    file_name: Optional[str] = "collection.json",
    config: Optional[Config] = None,
    messages: Optional[MessageQueue] = None,
    clients: Optional[List[Client]] = None,
    keep_non_downloaded: bool = False,
) -> Collection:
    """Downloads a collection to the local filesystem, synchronously.

    Does not download the collection's items' assets -- use
    :py:func:`download_item_collection` to download multiple items.

    Args:
        collection: A pystac collection
        directory: The destination directory
        file_name: The name of the collection file to save. If not provided,
            will not be saved.
        config: The download configuration
        messages: An optional queue to use for progress reporting
        clients: Pre-configured clients to use for access
        keep_non_downloaded: Keep all assets on the item, even if they're not
            downloaded.

    Returns:
        Collection: The collection, with updated asset hrefs

    Raises:
        CantIncludeAndExclude: Raised if both include and exclude are not None.
    """
    return asyncio.run(
        _functions.download_collection(
            collection=collection,
            directory=directory,
            file_name=file_name,
            config=config,
            messages=messages,
            clients=clients,
            keep_non_downloaded=keep_non_downloaded,
        )
    )


def download_item_collection(
    item_collection: ItemCollection,
    directory: PathLikeObject,
    path_template: Optional[str] = None,
    file_name: Optional[str] = "item-collection.json",
    config: Optional[Config] = None,
    messages: Optional[MessageQueue] = None,
    clients: Optional[List[Client]] = None,
    keep_non_downloaded: bool = False,
) -> ItemCollection:
    """Downloads an item collection to the local filesystem, synchronously.

    Args:
        item_collection: The item collection to download
        directory: The destination directory
        path_template: String to be interpolated to specify where to store
            downloaded files.
        file_name: The name of the item collection file to save. If not
            provided, will not be saved.
        config: The download configuration
        messages: An optional queue to use for progress reporting
        clients: Pre-configured clients to use for access
        keep_non_downloaded: Keep all assets on the item, even if they're not
            downloaded.

    Returns:
        ItemCollection: The item collection, with updated asset hrefs

    Raises:
        CantIncludeAndExclude: Raised if both include and exclude are not None.
    """
    return asyncio.run(
        _functions.download_item_collection(
            item_collection=item_collection,
            directory=directory,
            file_name=file_name,
            config=config,
            messages=messages,
            clients=clients,
            keep_non_downloaded=keep_non_downloaded,
            path_template=path_template,
        )
    )


def download_asset(
    key: str,
    asset: Asset,
    path: Path,
    config: Config,
    messages: Optional[MessageQueue] = None,
    clients: Optional[Clients] = None,
) -> Asset:
    """Downloads an asset, synchronously.

    Args:
        key: The asset key
        asset: The asset
        path: The path to which the asset will be downloaded
        config: The download configuration
        messages: An optional queue to use for progress reporting
        clients: A async-safe cache of clients. If not provided, a new one
            will be created.

    Returns:
        Asset: The asset with an updated href

    Raises:
        ValueError: Raised if the asset does not have an absolute href
    """
    return asyncio.run(
        _functions.download_asset(
            key=key,
            asset=asset,
            path=path,
            config=config,
            messages=messages,
            clients=clients,
        )
    )


def assert_asset_exists(
    asset: Asset,
    config: Optional[Config] = None,
    clients: Optional[List[Client]] = None,
) -> None:
    """Asserts that an asset exists, synchronously.

    Raises the source error if it does not.

    Args:
        asset: The asset the check for existence
        config: The download configuration to use for the existence check
        clients: Any pre-configured clients to use for the existence check

    Raises:
        Exception: An exception from the underlying client.
    """
    asyncio.run(_functions.assert_asset_exists(asset, config, clients))


def asset_exists(
    asset: Asset,
    config: Optional[Config] = None,
    clients: Optional[List[Client]] = None,
) -> bool:
    """Returns true if an asset exists, synchronously.

    Args:
        asset: The asset the check for existence
        config: The download configuration to use for the existence check
        clients: Any pre-configured clients to use for the existence check

    Returns:
        bool: Whether the asset exists or not
    """
    return asyncio.run(_functions.asset_exists(asset, config, clients))


def read_href(
    href: str, config: Optional[Config] = None, clients: Optional[List[Client]] = None
) -> bytes:
    """Reads an href and returns its bytes.

    Args:
        href: The href to read
        config: The download configuration to use
        clients: Any pre-configured clients to use

    Returns:
        bytes: The bytes from the href
    """
    return asyncio.run(_functions.read_href(href, config, clients))


def download_file(
    href: str,
    destination: PathLikeObject,
    config: Optional[Config] = None,
    clients: Optional[List[Client]] = None,
) -> None:
    """Downloads a file collection to the local filesystem.

    Args:
        href: The source href
        destination: The destination file path
        config: The download configuration
        clients: Pre-configured clients to use for access
    """
    return asyncio.run(_functions.download_file(href, destination, config, clients))



"""Read and download STAC items, item collections, collections, and assets.

The main entry points are free functions:

* :py:func:`download_item`
* :py:func:`download_item_collection`
* :py:func:`download_collection`
* :py:func:`download_asset`

Use :py:class:`Config` to configure how assets are downloaded.  Every client
inherits from :py:class:`Client`, which defines a common interface for accessing
assets.  Writing items, item collections, collections, and assets is currently
unsupported, but is on the roadmap.
"""

from ._functions import (
    assert_asset_exists,
    asset_exists,
    download_asset,
    download_collection,
    download_file,
    download_item,
    download_item_collection,
    open_href,
    read_href,
)
from .client import Client
from .config import Config
from .earthdata_client import EarthdataClient
from .errors import (
    AssetOverwriteError,
    ConfigError,
    ContentTypeError,
    DownloadError,
    DownloadWarning,
)
from .filesystem_client import FilesystemClient
from .http_client import HttpClient
from .messages import Message
from .planetary_computer_client import PlanetaryComputerClient
from .s3_client import S3Client
from .strategy import ErrorStrategy, FileNameStrategy

# Keep this list sorted
__all__ = [
    "AssetOverwriteError",
    "Client",
    "Config",
    "ConfigError",
    "ContentTypeError",
    "DownloadError",
    "DownloadWarning",
    "EarthdataClient",
    "ErrorStrategy",
    "FileNameStrategy",
    "FilesystemClient",
    "HttpClient",
    "Message",
    "PlanetaryComputerClient",
    "S3Client",
    "assert_asset_exists",
    "asset_exists",
    "download_asset",
    "download_collection",
    "download_item",
    "download_item_collection",
    "download_file",
    "open_href",
    "read_href",
]



from typing import Any, List


class AssetOverwriteError(Exception):
    """Raised when an asset would be overwritten during download."""

    def __init__(self, hrefs: List[str]) -> None:
        super().__init__(
            f"assets have the same file names and would overwrite each other: {hrefs}"
        )


class DownloadWarning(Warning):
    """A warning for when something couldn't be downloaded.

    Used when we don't want to cancel all downloads, but still inform the user
    about the problem.
    """


class ConfigError(Exception):
    """Raised if the configuration is not valid."""


class ContentTypeError(Exception):
    """The expected content type does not match the actual content type."""

    def __init__(self, actual: str, expected: str, *args: Any, **kwargs: Any) -> None:
        super().__init__(
            f"the actual content type does not match the expected: actual={actual}, "
            f"expected={expected}",
            *args,
            **kwargs,
        )


class DownloadError(Exception):
    """A collection of exceptions encountered while downloading."""

    exceptions: List[Exception]

    def __init__(self, exceptions: List[Exception], *args: Any, **kwargs: Any) -> None:
        self.exceptions = exceptions
        messages = list()
        for exception in exceptions:
            messages.append(f"{type(exception).__name__}: {exception}")
        super().__init__("\n".join(messages), *args, **kwargs)



import pytest
from stac_asset import Config, ConfigError


def test_validate_default() -> None:
    config = Config()
    config.validate()


def test_validate_include_and_exclude() -> None:
    config = Config(include=["foo"], exclude=["bar"])
    with pytest.raises(ConfigError):
        config.validate()


def test_warn_and_fail_fast() -> None:
    config = Config(warn=True, fail_fast=True)
    with pytest.raises(ConfigError):
        config.validate()



import pytest
from stac_asset import Config, HttpClient

pytestmark = [
    pytest.mark.asyncio,
]


@pytest.mark.vcr
async def test_href_exists() -> None:
    async with await HttpClient.from_config(Config()) as client:
        assert await client.href_exists("https://github.com/stac-utils/stac-asset")


async def test_default_http_timeout() -> None:
    async with await HttpClient.from_config(Config(http_client_timeout=42)) as client:
        assert client.session.timeout.total == 42



import os.path
from pathlib import Path

import pytest
from stac_asset import Config, EarthdataClient

pytestmark = [
    pytest.mark.skipif(
        os.environ.get("EARTHDATA_PAT") is None,
        reason="EARTHDATA_PAT is not set",
    ),
    pytest.mark.asyncio,
    pytest.mark.network_access,
]


async def test_download_href(tmp_path: Path) -> None:
    href = "https://data.lpdaac.earthdatacloud.nasa.gov/lp-prod-protected/MYD11A1.061/MYD11A1.A2023145.h14v17.061.2023146183035/MYD11A1.A2023145.h14v17.061.2023146183035.hdf"
    async with await EarthdataClient.from_config(Config()) as client:
        await client.download_href(href, tmp_path / "out.hdf")
        assert os.path.getsize(tmp_path / "out.hdf") == 197419



from __future__ import annotations

import copy
from dataclasses import dataclass, field
from typing import Dict, List, Optional

from .errors import ConfigError
from .strategy import ErrorStrategy, FileNameStrategy

DEFAULT_S3_REGION_NAME = "us-west-2"
DEFAULT_S3_RETRY_MODE = "adaptive"
DEFAULT_S3_MAX_ATTEMPTS = 10
DEFAULT_HTTP_CLIENT_TIMEOUT = 300


@dataclass
class Config:
    """Configuration for downloading items and their assets."""

    alternate_assets: List[str] = field(default_factory=list)
    """Alternate asset keys to prefer, if available."""

    file_name_strategy: FileNameStrategy = FileNameStrategy.FILE_NAME
    """The file name strategy to use when downloading assets."""

    warn: bool = False
    """If an error occurs during download, warn instead of raising the error."""

    fail_fast: bool = False
    """If an error occurs during download, fail immediately.

    By default, all downloads are completed before raising/warning any errors.
    Mutually exclusive with ``warn``.
    """

    error_strategy: ErrorStrategy = ErrorStrategy.DELETE
    """The strategy to use when errors occur during download."""

    exclude: List[str] = field(default_factory=list)
    """Assets to exclude from the download.

    Mutually exclusive with ``include``.
    """

    include: List[str] = field(default_factory=list)
    """Assets to include in the download.

    Mutually exclusive with ``exclude``.
    """

    make_directory: bool = True
    """Whether to create the output directory.

    If False, and the output directory does not exist, an error will be raised.
    """

    clean: bool = True
    """If true, clean up the downloaded file if it errors."""

    overwrite: bool = False
    """Download files even if they already exist locally."""

    http_client_timeout: Optional[float] = DEFAULT_HTTP_CLIENT_TIMEOUT
    """Total number of seconds for the whole request."""

    http_check_content_type: bool = True
    """If true, check the asset's content type against the response from the server."""

    http_headers: Dict[str, str] = field(default_factory=dict)
    """Extra headers to include in every http request."""

    earthdata_token: Optional[str] = None
    """A token for logging in to Earthdata."""

    s3_region_name: str = DEFAULT_S3_REGION_NAME
    """Default s3 region."""

    s3_requester_pays: bool = False
    """If using the s3 client, enable requester pays."""

    s3_retry_mode: str = DEFAULT_S3_RETRY_MODE
    """The retry mode to use for s3 requests."""

    s3_max_attempts: int = DEFAULT_S3_MAX_ATTEMPTS
    """The maximum number of attempts when downloading assets from s3."""

    s3_endpoint_url: Optional[str] = None
    """Set an optional custom endpoint url for s3."""

    def validate(self) -> None:
        """Validates this configuration.

        Raises:
            CannotIncludeAndExclude: ``include`` and ``exclude`` are mutually exclusive
        """
        if self.include and self.exclude:
            raise ConfigError(
                f"cannot provide both include and exclude: include={self.include}, "
                "exclude={self.exclude}"
            )
        if self.warn and self.fail_fast:
            raise ConfigError("cannot warn and fail fast as the same time")

    def copy(self) -> Config:
        """Returns a deep copy of this config.

        Returns:
            Config: A deep copy of this config.
        """
        return copy.deepcopy(self)



import json
from pathlib import Path

import stac_asset.blocking
from pystac import Collection, Item, ItemCollection
from stac_asset import Config


def test_download_item(tmp_path: Path, item: Item) -> None:
    item = stac_asset.blocking.download_item(item, tmp_path)
    item.validate()


def test_download_collection(tmp_path: Path, collection: Collection) -> None:
    collection = stac_asset.blocking.download_collection(collection, tmp_path)
    collection.validate()


def test_download_item_collection(
    tmp_path: Path, item_collection: ItemCollection
) -> None:
    item_collection = stac_asset.blocking.download_item_collection(
        item_collection, tmp_path
    )
    for item in item_collection:
        item.validate()


def test_download_asset(tmp_path: Path, item: Item) -> None:
    asset = stac_asset.blocking.download_asset(
        "data", item.assets["data"], tmp_path / "image.jpg", Config()
    )
    assert asset.href == str(tmp_path / "image.jpg")


def test_assert_asset_exists(item: Item) -> None:
    stac_asset.blocking.assert_asset_exists(item.assets["data"])


def test_asset_exists(item: Item) -> None:
    assert stac_asset.blocking.asset_exists(item.assets["data"])


def test_read_href(data_path: Path) -> None:
    text = stac_asset.blocking.read_href(str(data_path / "item.json"))
    Item.from_dict(json.loads(text))


def test_download_file(data_path: Path, tmp_path: Path) -> None:
    stac_asset.blocking.download_file(
        str(data_path / "item.json"), tmp_path / "item.json"
    )
    Item.from_file(tmp_path / "item.json")



from __future__ import annotations

from abc import ABC, abstractmethod
from asyncio import Lock, QueueFull
from pathlib import Path
from types import TracebackType
from typing import AsyncIterator, Dict, List, Optional, Type, TypeVar

import aiofiles
from yarl import URL

from .config import Config
from .messages import (
    WriteChunk,
)
from .types import MessageQueue, PathLikeObject

T = TypeVar("T", bound="Client")


class Client(ABC):
    """An abstract base class for all clients."""

    @classmethod
    async def from_config(cls: Type[T], config: Config) -> T:
        """Creates a client using the provided configuration.

        Needed because some client setups require async operations.

        Returns:
            T: A new client Client
        """
        return cls()

    def __init__(self) -> None:
        pass

    @abstractmethod
    async def open_url(
        self,
        url: URL,
        content_type: Optional[str] = None,
        messages: Optional[MessageQueue] = None,
    ) -> AsyncIterator[bytes]:
        """Opens a url and yields an iterator over its bytes.

        This is the core method that all clients must implement.

        Args:
            url: The input url
            content_type: The expected content type, to be checked by the client
                implementations
            messages: An optional queue to use for progress reporting

        Yields:
            AsyncIterator[bytes]: An iterator over chunks of the read file
        """
        # https://github.com/python/mypy/issues/5070
        if False:  # pragma: no cover
            yield

    async def open_href(
        self,
        href: str,
        content_type: Optional[str] = None,
        messages: Optional[MessageQueue] = None,
    ) -> AsyncIterator[bytes]:
        """Opens a href and yields an iterator over its bytes.

        Args:
            href: The input href
            content_type: The expected content type
            messages: An optional queue to use for progress reporting

        Yields:
            AsyncIterator[bytes]: An iterator over chunks of the read file
        """
        async for chunk in self.open_url(
            URL(href), content_type=content_type, messages=messages
        ):
            yield chunk

    async def download_href(
        self,
        href: str,
        path: PathLikeObject,
        clean: bool = True,
        content_type: Optional[str] = None,
        messages: Optional[MessageQueue] = None,
    ) -> None:
        """Downloads a file to the local filesystem.

        Args:
            href: The input href
            path: The output file path
            clean: If an error occurs, delete the output file if it exists
            content_type: The expected content type
            messages: An optional queue to use for progress reporting
        """
        try:
            async with aiofiles.open(path, mode="wb") as f:
                async for chunk in self.open_href(
                    href, content_type=content_type, messages=messages
                ):
                    await f.write(chunk)
                    if messages:
                        try:
                            messages.put_nowait(
                                WriteChunk(href=href, path=Path(path), size=len(chunk))
                            )
                        except QueueFull:
                            pass
        except Exception as err:
            path_as_path = Path(path)
            if clean and path_as_path.exists():
                try:
                    path_as_path.unlink()
                except Exception:
                    pass
            raise err

    async def href_exists(self, href: str) -> bool:
        """Returns true if the href exists.

        The default implementation naïvely opens the href and reads one chunk.
        Clients may implement specialized behavior.

        Args:
            href: The href to open

        Returns:
            bool: Whether the href exists
        """
        try:
            await self.assert_href_exists(href)
        except Exception:
            return False
        else:
            return True

    async def assert_href_exists(self, href: str) -> None:
        """Asserts that a href exists.

        The default implementation naïvely opens the href and reads one chunk.
        Clients may implement specialized behavior.

        Args:
            href: The href to open

        Raises:
            Exception: The underlying error when trying to open the file.
        """
        async for _ in self.open_href(href):
            break

    async def close(self) -> None:
        """Close this client."""
        pass

    async def __aenter__(self) -> Client:
        return self

    async def __aexit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Optional[TracebackType],
    ) -> Optional[bool]:
        return None


class Clients:
    """An async-safe cache of clients."""

    lock: Lock
    clients: Dict[Type[Client], Client]
    config: Config

    def __init__(self, config: Config, clients: Optional[List[Client]] = None) -> None:
        self.lock = Lock()
        self.clients = dict()
        if clients:
            # TODO check for duplicate types in clients list
            for client in clients:
                self.clients[type(client)] = client
        self.config = config

    async def get_client(self, href: str) -> Client:
        """Gets a client for the provided href.

        Args:
            href: The file href to download

        Returns:
            Client: An instance of that client.
        """
        from .filesystem_client import FilesystemClient
        from .http_client import HttpClient
        from .planetary_computer_client import PlanetaryComputerClient
        from .s3_client import S3Client

        url = URL(href)
        if not url.host:
            client_class: Type[Client] = FilesystemClient
        elif url.scheme == "s3":
            client_class = S3Client
        elif url.host.endswith("blob.core.windows.net"):
            client_class = PlanetaryComputerClient
        elif url.scheme == "http" or url.scheme == "https":
            client_class = HttpClient
        else:
            raise ValueError(f"could not guess client class for href: {href}")

        async with self.lock:
            if client_class in self.clients:
                return self.clients[client_class]
            else:
                client = await client_class.from_config(self.config)
                self.clients[client_class] = client
                return client

    async def close_all(self) -> None:
        """Close all clients."""
        async with self.lock:
            for client in self.clients.values():
                await client.close()



from .errors import ContentTypeError

ALLOWABLE_PAIRS = [
    ("image/tiff", "image/tiff; application=geotiff; profile=cloud-optimized"),
    ("text/xml", "application/xml"),
]
IGNORED_CONTENT_TYPES = ["binary/octet-stream", "application/octet-stream"]


def content_type(actual: str, expected: str) -> None:
    """Validates that the actual content type matches the expected.

    This is normally a simple string comparison, but has some extra rules:

    * COGs are allowed in place of TIFFs, and vice versa
    * Responses with ``binary/octet-stream`` and ``application/octet-stream``
      are always allowed

    Args:
        actual: The actual content type
        expected: The expected content type

    Raises:
        ContentTypeError: Raised if the actual doesn't match the expected.
    """
    if (
        actual != expected
        and actual not in IGNORED_CONTENT_TYPES
        and (actual, expected) not in ALLOWABLE_PAIRS
        and (expected, actual) not in ALLOWABLE_PAIRS
    ):
        raise ContentTypeError(actual=actual, expected=expected)



from __future__ import annotations

import os.path
from types import TracebackType
from typing import AsyncIterator, Optional, Type

import aiofiles
from yarl import URL

from .client import Client
from .messages import OpenUrl
from .types import MessageQueue


class FilesystemClient(Client):
    """A simple client for moving files around on the filesystem.

    Mostly used for testing, but could be useful in some real-world cases.
    """

    async def open_url(
        self,
        url: URL,
        content_type: Optional[str] = None,
        messages: Optional[MessageQueue] = None,
    ) -> AsyncIterator[bytes]:
        """Iterates over data from a local url.

        Args:
            url: The url to read bytes from
            content_type: The expected content type. Ignored by this client,
                because filesystems don't have content types.
            messages: An optional queue to use for progress reporting

        Yields:
            AsyncIterator[bytes]: An iterator over the file's bytes.

        Raises:
            ValueError: Raised if the url has a scheme. This behavior will
                change if/when we support Windows paths.
        """
        if url.scheme:
            raise ValueError(
                "cannot read a file with the filesystem client if it has a url scheme: "
                + str(url)
            )
        if messages:
            await messages.put(OpenUrl(size=os.path.getsize(url.path), url=url))
        async with aiofiles.open(url.path, "rb") as f:
            async for chunk in f:
                yield chunk

    async def assert_href_exists(self, href: str) -> None:
        """Asserts that an href exists."""
        if not os.path.exists(href):
            raise ValueError(f"href does not exist on the filesystem: {href}")

    async def __aenter__(self) -> FilesystemClient:
        return self

    async def __aexit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Optional[TracebackType],
    ) -> Optional[bool]:
        return None
